/* 
	 
  Sbzr, a program for editing bzr fonts
	Copyright (C) 1994 Serge Rosmorduc


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

      This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

      You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

THANKS : This package had a forerunner, with whom it doesn't share a
line of code (mainly because it was easier to do everything from
scatch with SUIT) Its forerunner was written by m. Louet and Corler,
using Xlib, as a computer science project. It allowed me to edit and
correct my fonts, and I want to thank them here for it.
*/

#include <stdio.h>
#include <strings.h>
#include <suit.h>
#include "bzr_opcodes.h"
#include <assert.h>
#include <unistd.h>
#include <values.h>
#include <math.h>
#include <sys/param.h>

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif

char disclaimer[]= "This program is free software; you can \
redistribute it and/or modify it under the terms of the GNU General \
Public License as published by the Free Software Foundation; either \
version 2 of the License, or (at your option) any later version. This \
program is distributed in the hope that it will be useful, but WITHOUT \
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or \
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License \
for more details. You should have received a copy of the GNU General \
Public License along with this program; if not, write to the Free \
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.";


char * etats_str[20]={
	"move point",
	"add point",
	"delete point",
	"delete contour",
	"add contour",
	"invert contour",
	"contour to top",
	"spline <-> line",
	"aligner splines",
	"move bounding box",
	"set rule origin",
	"set rule ticks",
	"select Contour",
	"move selected",
	"rotate selected",
	"scale selected",
	"0"
};


enum etat {
	movepoint, 
	addpoint, 
	deletepoint, 
	deletecontour,
	addcontour, 
	invertcontour,
	contour2top,
	spline2line,
	alignerspline,
	bbmove,
	setruleorigin,
	setruleticks,
	selectcontour,
	moveselected,
	rotateselected,
	scaleselected
} MonEtat;



#define MAXPTS 16
#define NEW(x) (x *)malloc(sizeof(x))
#define DIST(a,b) (((a).x-(b).x)*((a).x-(b).x) + \
((a).y-(b).y)*((a).y-(b).y))

unsigned char lastchar;

/*  creuser : 
	 dans Trait, un Trait ** qui pointe vers debut
*/

typedef struct Trait_struct{
	enum {ligne, spline} type;
	int selected;
	GP_point controle1, controle2;
	GP_point dernier;
	GP_point trace[MAXPTS];				/* le trace est garde en memoire */
	struct Trait_struct 
		* next, * previous;
} Trait;

typedef struct Contour_struct {
	enum senscontour {positif, negatif, cercle, droite } type;
	int selected;
	GP_point start;
	Trait * debut;
	struct Contour_struct 
		*next, *previous;
} Contour;

typedef struct Lettre_struct {
	unsigned char code;
	int used;
	double width;
	double llx, lly, urx, ury;
	Contour * contours;
} Lettre;

/* structures for storing fonts */
#define CHRINFONT 256

typedef struct Font_struct {
	char * name;
	char * filename;
	double size;
	Lettre lettres[CHRINFONT];
} Font;


typedef unsigned char Breel[3];
#define MAXREEL (255.0 + 255.0/256 + 255.0/(256*256))

SUIT_object LaVisu, Affiche;
Font * LaFonte = NULL;
Font defaultfont;
int Zoom= FALSE;

GP_point dragPoint;
GP_rectangle dragRectangle;

Contour *clip= NULL;

enum {horizontal, vertical} direction_mod_bb;

double read_Bdouble(FILE * f)
{
	double d;
	unsigned char c[3];
	fread(c, sizeof(char), 3, f);
	d= c[0] + (double)c[1]/ (1 << 8) + (double)c[2]/ (1 << 16);
	if (c[0] & (1 << 7))
		d= d - MAXREEL;
	return d;
}

void write_Bdouble(FILE *f, double d)
{
	unsigned char c[3];
	if (d < 0)
		d= d+ MAXREEL;
	c[0]= floor(d);
	d= (d-floor(d))*256;
	c[1]= floor(d);
	d= (d-floor(d))*256;
	c[2]= floor(d);
	fwrite(c, sizeof(char), 3,f);
}

double read_Sdouble(FILE * f)
{
	double d;
	unsigned char c[2];
	fread(c, sizeof(char), 2, f);
	d= (double)c[0]/ (1 << 8) + (double)c[1]/ (1 << 16);
	if (c[0] & (1 << 7))
		d= d - 1;
	return d;
}

void fixe_sens_contour(Contour *cont)
{
	double val= 0.0;
	GP_point prec;
	Trait *trace;
	trace= cont->debut;
	prec= cont->start;
	while (trace != NULL)
	{
		switch (trace->type)
		{
			int i;
		case ligne :
			val+= (trace->dernier.x -prec.x) * 
				(trace->dernier.y +prec.y);
			/* la division par 2.0 est inutile.*/
			break;
		case spline :
			for(i=0; i<MAXPTS; i++)
			{
				val+= (trace->trace[i].x -prec.x) * 
					(trace->trace[i].y +prec.y);
				prec= trace->trace[i];
			}
			val+= (trace->dernier.x -prec.x) * 
				(trace->dernier.y +prec.y);
			break;
		}
		prec= trace->dernier;
		trace= trace->next;
	}
	val+= (cont->start.x -prec.x) * 
		(cont->start.y -prec.y);
	if (val > 0.0)
		cont->type= negatif;
	else 
		cont->type= positif;
} 

void fixe_trace(GP_point depart,Trait *tr)
{
	double ux, uy,vx, vy ;
	double wx, wy, zx, zy;
	double t;
	int i;
	if (tr->type== ligne) 
		return;
	ux= - depart.x + 3*tr->controle1.x
		- 3*tr->controle2.x + tr->dernier.x;
	vx= 3*tr->controle2.x
		 + 3*depart.x -6 * tr->controle1.x;
	wx= 3*tr->controle1.x- 3*depart.x;
	zx= depart.x;
	uy= - depart.y + 3*tr->controle1.y
		- 3*tr->controle2.y + tr->dernier.y;
	vy= 3*tr->controle2.y
		 + 3*depart.y -6 * tr->controle1.y;
	wy= 3*tr->controle1.y- 3*depart.y;
	zy= depart.y;

	for (i= 0, t= 0.0; i<MAXPTS; i++)
	{
		t+= 1.0/(MAXPTS+1);
		tr->trace[i].x= ((ux*t+vx)*t+wx)*t+zx;
		tr->trace[i].y= ((uy*t+vy)*t+wy)*t+zy;
	}
}

Trait * select_line(SUIT_object visu, GP_point where, Contour **moncont)
{
	int car;
	double min;
	Contour *c;
	Trait * result;
	Lettre *l;
	
	car= SUIT_getInteger(visu, "char");
	result= NULL;
	*moncont= NULL;
	min= MAXDOUBLE;
	
	l= LaFonte->lettres + car;
	if (l->used)
	{
		*moncont= c= l->contours;
		
		while (c != NULL)
		{
			GP_point prev= c->start;
			Trait * current= c->debut;
			while (current !=NULL)
			{
				double dist;
				GP_point a,b;
				switch (current->type)
				{
				case ligne:
					a= prev;
					b= current->dernier;
					break;
				case spline :
					a= current->controle1;
					b= current->controle2;
					break;
				}

				if ( /*  les produits scalaires AW.AB et BW.BA sont > 0*/
						(a.x != b.x || a.y != b.y)
						&&
						((where.x- a.x)*(b.x - a.x) + (where.y- a.y)*(b.y - a.y)) >= 0.0
						&&
						((where.x- b.x)*(a.x - b.x) + (where.y- b.y)*(a.y - b.y)) >= 0.0)
				{
					dist= ((where.x- a.x)* (b.y - a.y)
								 -(where.y -a.y)* (b.x - a.x));
					dist= dist*dist/ DIST(a,b);
					/* pour ne pas avoir a utiliser sqrt*/ 
				}
				else
				{
					double d1, d2;
					d1= DIST(where,a);
					d2= DIST(where,b);
					dist = (d1<d2)?d1:d2;
				}
				if (dist < min)
				{
					min= dist;
					*moncont= c;
					result= current;
				}
				prev= current->dernier;
				current= current->next;
			}
			c= c->next;
		}
		if (result)
			result->selected= !(result->selected);
	}
	return result;
}




#define ESS_POINT(xx,ttt)\
if ((dist= DIST(where,(xx))) <= min)\
{\
	 *Po= (xx);\
	 *Tr= current;\
	 *Co= c;\
	 min= dist;\
	 *type= (ttt);\
}

enum typepoint {sommet, controle1, controle2};

int select_point(SUIT_object visu, GP_point where,
									Contour **Co, Trait **Tr, GP_point* Po, 
								 enum typepoint *type)
{
	int car, result;
	double min;
	Contour *c;
	Lettre *l;

	car= SUIT_getInteger(visu, "char");
	*Tr= NULL;
	*Co= NULL;
	*type= sommet;
	result= FALSE;

	min= MAXDOUBLE;
	
	l= LaFonte->lettres + car;
	if (l->used)
	{
		c= l->contours;
		
		while (c != NULL)
		{
			Trait * current= c->debut;
			result= TRUE;

			if (!*Co && !current)
			{
				*Co= c;
				*Po= c->start;
				*type= sommet;
				min= DIST(where,*Po);
			}
			
			while (current !=NULL)
			{
				switch (current->type)
				{
					double dist;
				case ligne:
					ESS_POINT(current->dernier,sommet);
					break;
				case spline :
					ESS_POINT(current->controle1,controle1);
					ESS_POINT(current->controle2,controle2);
					ESS_POINT(current->dernier,sommet);
					break;
				}
				current= current->next;
			}
			c= c->next;
		}
	}
	return result;
}

GP_point ArronditPoint(GP_point p)
{
	GP_point m;
	GP_point origine, tickpoint;
	GP_point u,v;
	double scal1, scal2;

	if (SUIT_getBoolean(LaVisu, "round to rule"))
	{
		origine.x= SUIT_getDouble(LaVisu, "rule origine x");
		origine.y= SUIT_getDouble(LaVisu, "rule origine y");
		tickpoint.x= SUIT_getDouble(LaVisu, "rule tick x");
		tickpoint.y= SUIT_getDouble(LaVisu, "rule tick y");
		u.x= tickpoint.x - origine.x;
		u.y= tickpoint.y - origine.y;
		v.x= -u.y;
		v.y= u.x;
		m.x= p.x- origine.x;
		m.y= p.y- origine.y;
		scal1= floor((u.x*m.x+u.y*m.y)/(u.x*u.x+u.y*u.y)+ 0.5);
		scal2= floor((v.x*m.x+v.y*m.y)/(v.x*v.x+v.y*v.y) +0.5);
		p.x= scal1* u.x+ scal2* v.x+ origine.x;
		p.y= scal1* u.y+ scal2* v.y +origine.y;
	}
	return p;
}

void SelectContour(SUIT_object visu, SUIT_event myEvent)
{
	Contour *c;
	Trait *t;
	GP_point p;
	enum typepoint type;

	select_point(visu,myEvent.worldLocation,
									&c,&t,&p, &type);
	if (c)
		c->selected= !c->selected;
}

void InvertContour(SUIT_object visu, SUIT_event myEvent)
{
	Contour *c;
	Trait *t;
	GP_point p;
	enum typepoint type;

	select_point(visu,myEvent.worldLocation,
									&c,&t,&p, &type);
	if (c)
		if(c->type==positif)
			c->type= negatif;
		else
			c->type= positif;
}


char skip_nop(FILE *f)
{
	int c;
	while ((c=getc(f)) == NO_OP);
	lastchar= c;
	return c;
}

Font * read_font(char *name)
{
	Font *fonte;
	FILE *f;
	int i, numchar;
	char c;
	double size;

	if ((f= fopen(name,"r")) == NULL)
		return NULL;

	fonte= NEW(Font);
	for (i=0; i< CHRINFONT; i++)
		fonte->lettres[i].used= FALSE;
	fonte->filename= strdup(name);
	size= fonte->size= read_Bdouble(f);
	numchar= (unsigned)getc(f);
	fonte->name= 
		(char *)malloc((numchar+1)*sizeof(char));
	fonte->name[numchar]=0;
	fread(fonte->name,sizeof(char),
				numchar,f);
	while ((c=skip_nop(f)) != POST)
	{
		Contour *lastcont, *currcont;
		Lettre *ltr;
		int code;
		code= (unsigned)getc(f);
		ltr= fonte->lettres + code;
		ltr->code= code;
		switch(c)
		{
		case BOC_ABBREV :
			ltr->used= TRUE;
			ltr->contours= NULL;
			ltr->width= read_Sdouble(f) * size;
			ltr->llx= read_Sdouble(f) * size;
			ltr->lly= read_Sdouble(f) * size;
			ltr->urx= read_Sdouble(f) * size;
			ltr->ury= read_Sdouble(f) * size;
			break;
		case BOC:
			ltr->used= TRUE;
			ltr->contours= NULL;
			ltr->width= read_Bdouble(f) * size;
			ltr->llx= read_Bdouble(f) * size;
			ltr->lly= read_Bdouble(f) * size;
			ltr->urx= read_Bdouble(f) * size;
			ltr->ury= read_Bdouble(f) * size;
			break;
		default:
			break;
		}
		/* paths */
		lastcont= NULL;
		c= skip_nop(f);
		while ( c != EOC)
		{
			Trait * lasttrait;
			GP_point lastpoint;
			assert(c== START_PATH);
			currcont= NEW(Contour);
			currcont->selected= FALSE;
			currcont->next= NULL;
			currcont->debut= NULL;
			currcont->start.x= read_Bdouble(f) * size;
			currcont->start.y= read_Bdouble(f) * size;
			lasttrait= NULL;
			lastpoint= currcont->start;
			while ((c= skip_nop(f)) != START_PATH &&
						 c != EOC)
			{
				if (lasttrait)
				{
					lasttrait->next= NEW(Trait);
					lasttrait->next->previous= lasttrait;
					lasttrait= lasttrait->next;
				}
				else
				{
					lasttrait= currcont->debut= NEW(Trait);
					lasttrait->previous= NULL;
				}
				lasttrait->next= NULL;
				lasttrait->selected= FALSE;

				switch(c)
				{
				case LINE:
					lasttrait->type= ligne;
					lasttrait->dernier.x= read_Bdouble(f) * size;
					lasttrait->dernier.y= read_Bdouble(f) * size;					
					break;
				case LINE_ABBREV:
					lasttrait->type= ligne;
					lasttrait->dernier.x= read_Sdouble(f) * size;
					lasttrait->dernier.y= read_Sdouble(f) * size;					
					break;
				case SPLINE:
					lasttrait->type= spline;
					lasttrait->controle1.x= read_Bdouble(f) * size;
					lasttrait->controle1.y= read_Bdouble(f) * size;
					lasttrait->controle2.x= read_Bdouble(f) * size;
					lasttrait->controle2.y= read_Bdouble(f) * size;
					lasttrait->dernier.x= read_Bdouble(f) * size;
					lasttrait->dernier.y= read_Bdouble(f) * size;
					break;
				case SPLINE_ABBREV:
					lasttrait->type= spline;
					lasttrait->controle1.x= read_Sdouble(f) * size;
					lasttrait->controle1.y= read_Sdouble(f) * size;
					lasttrait->controle2.x= read_Sdouble(f) * size;
					lasttrait->controle2.y= read_Sdouble(f) * size;
					lasttrait->dernier.x= read_Sdouble(f) * size;
					lasttrait->dernier.y= read_Sdouble(f) * size;
					break;
				}
				fixe_trace(lastpoint, lasttrait);
				lastpoint= lasttrait->dernier;
			}

			if (currcont->debut)
			{
				fixe_sens_contour(currcont);
				if (lastcont)
				{
					lastcont->next= currcont;
					currcont->previous= lastcont;
					lastcont= currcont;
				}
				else
				{
					lastcont= ltr->contours= currcont;
					lastcont->previous= NULL;
				}
			}
			else
			{
				free(currcont);					/* pas de contour sans trait chez moi ! */
			}
		}
	}
	fclose(f);
	return fonte;
}

/** Inverse Le Contour *c */
void repare_sens(Contour **c)
{
	Trait *t, *lastt;
	lastt= NULL;
	if ((t= (*c)->debut))
	{
		GP_point prev;
		prev= (*c)->start;
		while(t)
		{
			Trait *tmp;
			GP_point tmppoint;
			tmp= t->next;
			t->next= t->previous;
			t->previous= tmp;
			tmppoint= prev;
			prev= t->dernier;
			t->dernier= tmppoint;
			switch(t->type)
			{
				GP_point tmppoint;
			case ligne:
				break;
			case spline:
				tmppoint= t->controle1;
				t->controle1= t->controle2;
				t->controle2= tmppoint;
				fixe_trace(prev,t);
				break;
			}
			lastt= t;
			t= t->previous;						/* ancient next ! */
		} 
		(*c)->debut= lastt;
	}
	if ((*c)->type== positif)
		(*c)->type= negatif;
	else
		(*c)->type= positif;
}
	
void write_fonte(Font* fo, char *fname)
{
	FILE *f;
	double size;
	double llx= MAXDOUBLE;
	double lly= MAXDOUBLE;
	double urx= -MAXDOUBLE;
	double ury= -MAXDOUBLE;
	unsigned int nbrchar=0;
	unsigned int where= 0;
	unsigned int postptr;
	unsigned int wherelettres[256];

	if ((f= fopen(fname, "w")))
	{
		unsigned int i;
		size= fo->size;
		write_Bdouble(f, fo->size);
		putc(strlen(fo->name),f);
		fputs(fo->name,f);
		where= 3 + 1+ strlen(fo->name);
		for (i= 0; i<256; i++)
		{
			Lettre *l;
			if (fo->lettres[i].used)
			{
				Contour *c;
				nbrchar++;
				wherelettres[i]= where;
				putc(BOC,f); where++;
				putc(i,f); where++;
				l= fo->lettres +i;
				write_Bdouble(f, l->width/size);
				write_Bdouble(f, l->llx/size);
				write_Bdouble(f, l->lly/size);
				write_Bdouble(f, l->urx/size);
				write_Bdouble(f, l->ury/size);
				if (llx > l->llx)
					llx=  l->llx;
				if (lly > l->lly)
					lly=  l->lly;
				if (urx < l->urx)
					urx=  l->urx;
				if (ury < l->ury)
					ury=  l->ury;
				where+= 5*3;
				c= l->contours;
				while (c)
				{
					Trait *t;
					enum senscontour sens;
					sens= c->type;
					fixe_sens_contour(c);
					if (sens != c->type)
						repare_sens(&c);
					putc(START_PATH,f); where++;
					write_Bdouble(f, c->start.x/size);				
					write_Bdouble(f, c->start.y/size);
					where+= 2*3;
					t= c->debut;
					while(t)
					{
						switch(t->type)
						{
						case ligne:
							putc(LINE,f); where++;
							write_Bdouble(f, t->dernier.x/size);
							write_Bdouble(f, t->dernier.y/size);
							where+= 2*3;
							break;
						case spline:
							putc(SPLINE,f); where++;
							write_Bdouble(f, t->controle1.x/size);
							write_Bdouble(f, t->controle1.y/size);
							write_Bdouble(f, t->controle2.x/size);
							write_Bdouble(f, t->controle2.y/size);
							write_Bdouble(f, t->dernier.x/size);
							write_Bdouble(f, t->dernier.y/size);
							where+= 6*3;
							break;
						}
						t= t->next;
					}
					c= c->next;
				}
				putc(EOC,f); where++;
			}
		}
		postptr= where;
		putc(POST,f); where++;
		write_Bdouble(f, llx/size);
		write_Bdouble(f, lly/size);
		write_Bdouble(f, urx/size);		
		write_Bdouble(f, ury/size);
		where+= 4*3;
		for (i=0; i<256; i++)
		{
			if (fo->lettres[i].used)
			{
				unsigned int ou= wherelettres[i];
				putc(CHAR_LOC,f); where++;
				putc(i,f); where++;
				putc(ou >> 8*3,f); where++;
				putc((ou & 0xFF0000)>> 8*2,f); where++;
				putc((ou & 0xFF00)>> 8,f); where++;
				putc((ou & 0xFF),f); where++;
			}
		}
		putc(nbrchar,f); where++;
		putc(POST_POST,f); where++;
		putc(postptr >> 8*3,f); where++;
		putc((postptr & 0xFF0000)>> 8*2,f); where++;
		putc((postptr & 0xFF00)>> 8,f); where++;
		putc((postptr & 0xFF),f); where++;
		putc(BZR_ID,f); where++;
		fclose(f);
	}
}

void DrawZoom(point p)
{
	double delta;
	GP_point pp= GP_unMapPoint(p);
	delta= fabs(pp.x-dragPoint.x);
	GP_rectangleCoord(dragPoint.x-delta,dragPoint.y-delta,
										dragPoint.x+delta,dragPoint.y+delta);
}

void DrawDrag(point p)
{
	GP_line(dragPoint,GP_unMapPoint(p));
}

void drag_vertical(point p)
{
	GP_point m;
	m= GP_unMapPoint(p);
	GP_lineCoord(m.x, dragPoint.x, m.x, dragPoint.y);
}

void drag_horizontal(point p)
{
	GP_point m;
	m= GP_unMapPoint(p);
	GP_lineCoord(dragPoint.x, m.y, dragPoint.y,m.y);
}

void BBMove(SUIT_object visu, SUIT_event myEvent)
{
	double d[4], dd;
	GP_point pt, lepoint;
	Lettre *l;
	int car;
	int i, r;
	car= SUIT_getInteger(visu, "char");
	l= LaFonte->lettres + car;
	
	pt= myEvent.worldLocation;
	d[0]= fabs(l->llx- pt.x);
	d[1]= fabs(l->ury- pt.y);	
	d[2]= fabs(l->urx- pt.x);
	d[3]= fabs(l->lly- pt.y);
	
	for(i=1, r=0, dd= d[0]; i<4; i++)
		if (d[i]<dd)
		{
			r= i;
			dd= d[i];
		}
	
	switch(r)
	{
	case 0:
	case 2:
		dragPoint.x= l->lly;
		dragPoint.y= l->ury;
		lepoint= ArronditPoint(GP_unMapPoint(SUIT_drag(drag_vertical)));
		if (r==0)
			l->llx= lepoint.x;
		else
			l->urx= lepoint.x;
		break;
	case 1:
	case 3:
		dragPoint.x= l->llx;
		dragPoint.y= l->urx;
		lepoint= ArronditPoint(GP_unMapPoint(SUIT_drag(drag_horizontal)));
		if (r==1)
			l->ury= lepoint.y;
		else
			l->lly= lepoint.y;
		break;
	}
	l->width= l->urx -l->llx;
}

void TranslateContour(Contour* c, GP_point v)
{
	Trait *t;
	GP_point prev;
	
	c->start.x+= v.x;
	c->start.y+= v.y;
	prev= c->start;
	
	t= c->debut;
	
	while(t)
	{
		switch(t->type)
		{
		case ligne: 
			break;
		case spline:
			t->controle1.x+= v.x;
			t->controle1.y+= v.y;
			t->controle2.x+= v.x;
			t->controle2.y+= v.y;
			break;
		}
		t->dernier.x+= v.x;
		t->dernier.y+= v.y;
		fixe_trace(prev,t);
		prev= t->dernier;
		t= t->next;
	}
}

void RotatePoint(GP_point *p, GP_point centre, GP_point nouvx)
{
	GP_point m;
	m.x= p->x - centre.x;
	m.y= p->y - centre.y;
	p->x= m.x*nouvx.x - m.y*nouvx.y +centre.x;
	p->y= m.x*nouvx.y + m.y*nouvx.x +centre.y;
}

void RotateContour(Contour* c, GP_point centre, GP_point nouvx)
{
	Trait *t;
	GP_point prev;
	
	RotatePoint(&(c->start),centre,nouvx);
	prev= c->start;
	
	t= c->debut;
	
	while(t)
	{
		switch(t->type)
		{
		case ligne: 
			break;
		case spline:
			RotatePoint(&(t->controle1),centre,nouvx);
			RotatePoint(&(t->controle2),centre,nouvx);
			break;
		}
		RotatePoint(&(t->dernier),centre,nouvx);
		fixe_trace(prev,t);
		prev= t->dernier;
		t= t->next;
	}
}

void MoveToZero()
{
	Contour *co;
	GP_point vecteur;
	int car;
	Lettre* l;
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	if (l->used)
	{
		vecteur.x= -(l->llx);
		vecteur.y= -(l->lly);
		co= l->contours;
		while (co)
		{
			TranslateContour(co,vecteur);
			co= co->next;
		}
		l->urx -= l->llx;
		l->ury -= l->lly;
		l->llx= 0.0;
		l->lly= 0.0;
		SUIT_redisplayRequired(LaVisu);
	}
}

#define METVALSAJOUR(A) if (urx < (A).x) urx= (A).x; \
if (ury < (A).y) ury= (A).y; \
if (llx > (A).x) llx= (A).x; \
if (lly > (A).y) lly= (A).y


GP_rectangle FindBB(boolean selectedOnly)
{
	int car;
	Lettre* l;
	GP_rectangle result;
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	if (l->used)
	{
		double llx, lly, urx, ury;
		Contour *c;
		llx= MAXDOUBLE;
		lly= MAXDOUBLE;
		urx= -MAXDOUBLE;
		ury=-MAXDOUBLE;
		c= l->contours;
		while (c)
		{
			if (! selectedOnly || c->selected)
			{
				Trait*t;
				/* pas la peine de tester start, puisque tout contour a au moins
					 un trait */
				t= c->debut;
				while (t)
				{
					switch (t->type)
					{
						int i;
					case ligne:
						METVALSAJOUR(t->dernier);
						break;
					case spline:
						for (i=0; i<MAXPTS; i++)
						{
							METVALSAJOUR(t->trace[i]);
						}
						METVALSAJOUR(t->dernier);					
					}
					t= t->next;
				}
			}
			c= c->next;
		}
		result.bottom_left.x= llx;
		result.bottom_left.y= lly;
		result.top_right.x= urx;
		result.top_right.y= ury;
	}
	return result;
}

void ComputeBB()
{
	int car;
	Lettre* l;
	GP_rectangle rect;
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	if (l->used)
	{
		rect= FindBB(FALSE);
		l->urx= rect.top_right.x;
		l->ury= rect.top_right.y;
		l->llx= rect.bottom_left.x;
		l->lly= rect.bottom_left.y;
		l->width= l->urx-l->llx;
		SUIT_redisplayRequired(LaVisu);
	}
}

void MoveSelected(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	GP_point lepoint, vecteur;
	int car;
	Lettre* l;
	
	car= SUIT_getInteger(visu, "char");
	l= LaFonte->lettres + car;
	
	dragPoint= myEvent.worldLocation;
	lepoint= ArronditPoint(GP_unMapPoint(SUIT_drag(DrawDrag)));
	vecteur.x= lepoint.x-dragPoint.x;
	vecteur.y= lepoint.y-dragPoint.y;
	co= l->contours;
	while (co)
	{
		if (co->selected)
		{
			TranslateContour(co,vecteur);
		}
		co= co->next;
	}
}

void RotateSelected(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	GP_point lepoint, vecteur;
	int car;
	Lettre* l;
	
	car= SUIT_getInteger(visu, "char");
	l= LaFonte->lettres + car;
	
	dragPoint= myEvent.worldLocation;
	lepoint= ArronditPoint(GP_unMapPoint(SUIT_drag(DrawDrag)));
	vecteur.x= lepoint.x-dragPoint.x;
	vecteur.y= lepoint.y-dragPoint.y;
	if ( vecteur.x != 0.0 || vecteur.y != 0.0 )
	{
		double d;
		d= sqrt(DIST(dragPoint, lepoint));
		vecteur.x= vecteur.x /d;
		vecteur.y= vecteur.y /d;		/* on normalise */
		co= l->contours;
		while (co)
		{
			if (co->selected)
			{
				RotateContour(co,dragPoint,vecteur);
			}
			co= co->next;
		}
	}
}


void ScalePoint(GP_point *pt, GP_point centre, double rapport)
{
	pt->x = (pt->x - centre.x) * rapport + centre.x;
	pt->y = (pt->y - centre.y) * rapport + centre.y;
} 

void ScaleSelected(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	GP_point lepoint;
	GP_rectangle bound;
	double olddist, newdist, rapport;
	int car;
	Lettre* l;
	
	car= SUIT_getInteger(visu, "char");
	l= LaFonte->lettres + car;
	
	bound= FindBB(TRUE);
	
	dragPoint.x= (bound.bottom_left.x + bound.top_right.x)/ 2.0;
	dragPoint.y= (bound.bottom_left.y + bound.top_right.y)/ 2.0;

	olddist= DIST(dragPoint, myEvent.worldLocation);
	lepoint= ArronditPoint(GP_unMapPoint(SUIT_drag(DrawZoom)));
	newdist= DIST(lepoint, dragPoint);
	
	if (olddist == 0.0 || newdist == 0.0)
		return;

	rapport= sqrt(newdist/olddist);

	co= l->contours;
	while (co)
	{
		if (co->selected)
		{
			Trait *t;
			GP_point prev;
			t= co->debut;
			ScalePoint(&(co->start),dragPoint,rapport);
			prev= co->start;

			while(t)
			{
				switch(t->type)
				{
				case ligne: 
					break;
				case spline:
					ScalePoint(&(t->controle1), dragPoint, rapport);
					ScalePoint(&(t->controle2), dragPoint, rapport);
					break;
				}
				ScalePoint(&(t->dernier), dragPoint, rapport);
				fixe_trace(prev,t);
				prev= t->dernier;
				t= t->next;
			}
			
		}
		co= co->next;
	}
}


void DetruireContours(Contour *c)
{
	Contour *n;
	while(c)
	{
		Trait *t;
		n=c->next;
		t= c->debut;
		while (t)
		{
			Trait *next;
			next= t->next;
			free(t);
			t= next;
		}
		free(c);
		c= c->next;
	}
}

void DeleteContour(SUIT_object visu, SUIT_event myEvent)
{
	Contour *c;
	Trait *t;
	GP_point p;
	enum typepoint type;
	
	select_point(visu,myEvent.worldLocation,
									&c,&t,&p, &type);
	if (!c) return;
	if (c->previous)
	{
		if (c->next)
		{
			c->next->previous= c->previous;
		}
		c->previous->next= c->next;
	}
	else
	{
		int car; 
		Lettre *l;
		car= SUIT_getInteger(LaVisu, "char");
		l= LaFonte->lettres + car;
		l->contours= c->next;
		if (c->next)
			c->next->previous= NULL;
	}
	t= c->debut;
	while(t)
	{
		Trait *n;
		n= t->next;
		free(t);
		t=n;
	}
	free(c);
}

void CopySelected()
{
	Contour *co, **vers, *prev;
	int car;
	Lettre* l;
	
	/* fin du clipboard */
	vers= &clip;
	prev= NULL;
	while (*vers)
	{
		prev= *vers;
		vers= &((*vers)->next);
	}
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	
	co= l->contours;
	while (co)
	{
		if (co->selected)
		{
			Trait *t, **newt, *prevt;
			*vers= NEW(Contour);
			**vers= *co;
			(*vers)->previous= prev;
			(*vers)->next= NULL;
			if (prev)
				prev->next= *vers;
			t= co->debut;
			newt= &((*vers)->debut);
			prevt= NULL;
			while(t)
			{
				*newt= NEW(Trait);
				**newt= *t;
				(*newt)->previous= prevt;
				prevt= (*newt);
				newt= &((*newt)->next);
				t=t->next;
			}
			prev= *vers;
			vers= &((*vers)->next);
		}
		co= co->next;
	}
}

void pasteClip()
{
	Contour **co;
	int car;
	Lettre* l;
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	co= &(l->contours);
	
	while (*co)
		co= &((*co)->next);
	
	*co= clip;
	clip= NULL;
}

#define PREVIOUS_POINT(xtr,xcont)\
(((xtr)->previous)?(xtr)->previous->dernier:(xcont)->start)

#define NEXT_TRAIT(xtr,xcont)\
(((xtr)->next)?(xtr)->next:(xcont)->debut)


void MovePoint(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	Trait *tr;
	GP_point lepoint;
	enum typepoint type;
	
	if (select_point(visu, myEvent.worldLocation,
									 &co, &tr, &dragPoint, &type))
	{
		lepoint= ArronditPoint(GP_unMapPoint(SUIT_drag(DrawDrag)));
		if (tr)
		{
			switch (type)
			{
			case sommet:
				tr->dernier= lepoint;
				if (tr->next == NULL)
					co->start= lepoint;
				fixe_trace(lepoint, NEXT_TRAIT(tr,co));
				break;
			case controle1:
				tr->controle1= lepoint;
				break;
			case controle2:
				tr->controle2= lepoint;
				break;
			}
			fixe_trace(PREVIOUS_POINT(tr,co), tr);
		}
		else 
		{
			co->start= lepoint;
		}
	}
}

#define MOYENNE(a,bb,cc) (a).x= ((bb).x + (cc).x)/2.0;\
(a).y= ((bb).y + (cc).y)/2.0

void AddPoint(SUIT_object visu, SUIT_event myEvent)
{
	Trait *t, *nouv;
	Contour *cont;
	if ((t=	select_line(visu, myEvent.worldLocation, &cont)))
	{
		GP_point prev;
		nouv= NEW(Trait);
		nouv->previous= t;
		nouv->next= t->next;
		if (t->next)
			t->next->previous= nouv;
		t->next= nouv;
		prev= PREVIOUS_POINT(t,cont);
		t->selected= FALSE;
		nouv->selected=FALSE;
		
		switch(t->type)
		{
			GP_point ma,mb,m0,m1,m2,m3;
		case ligne:
			nouv->type= ligne;
			nouv->dernier=t->dernier;
			t->dernier.x= (nouv->dernier.x + prev.x)/2.0;
			t->dernier.y= (nouv->dernier.y + prev.y)/2.0;
			break;
		case spline:
			nouv->type= spline;
			MOYENNE(ma,prev,t->controle1);
			MOYENNE(mb,t->dernier,t->controle2);
			MOYENNE(m0,t->controle1,t->controle2);
			MOYENNE(m1,ma,m0);
			MOYENNE(m2,mb,m0);
			MOYENNE(m3,m1,m2);
			nouv->controle1= m2;
			nouv->controle2= mb;
			nouv->dernier= t->dernier;
			fixe_trace(m3,nouv);			
			t->controle1= ma;
			t->controle2= m1;
			t->dernier= m3;
			fixe_trace(prev,t);
			break;
		}
	}
}

void Spline2Line(SUIT_object visu, SUIT_event myEvent)
{
	Trait *t;
	Contour *cont;
	if ((t=	select_line(visu, myEvent.worldLocation, &cont)))
	{
		GP_point prev;
		prev= PREVIOUS_POINT(t,cont);
		t->selected= FALSE;
		
		switch(t->type)
		{
		case ligne:
			t->controle1.x= prev.x+0.25*(t->dernier.x-prev.x);
			t->controle1.y= prev.y+0.25*(t->dernier.y-prev.y);
			t->controle2.x= t->dernier.x-0.25*(t->dernier.x-prev.x);
			t->controle2.y= t->dernier.y-0.25*(t->dernier.y-prev.y);
			
			t->type= spline;
			fixe_trace(prev,t);
			break;
		case spline:
			t->type= ligne;
			fixe_trace(prev,t);
			break;
		}
	}
}

void AddContour(SUIT_object visu, SUIT_event myEvent)
{
	int car;
	Lettre *l;
	Contour *c, *lc, *prev;
	Trait *t;
	
	car= SUIT_getInteger(visu, "char");
	l= LaFonte->lettres + car;
	c= NEW(Contour);
	c->selected= FALSE;
	t= NEW(Trait);
	
	c->type= positif;
	c->start= ArronditPoint(myEvent.worldLocation);
	c->debut= t;
	c->next= NULL;
	c->previous= NULL;
	
	t->type= ligne;
	t->selected= FALSE;
	t->dernier= ArronditPoint(myEvent.worldLocation);
	t->next= t->previous= NULL;
	
	prev= NULL;
	lc= l->contours;
	while (lc)
	{
		prev= lc;
		lc= lc->next;
	}
	if (prev)
	{
		prev->next= c;
		c->previous= prev;
	}
	else
	{
		l->contours= c;
	}
}

void DetruirePoint(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	Trait *tr;
	enum typepoint type;
	
	if (select_point(visu, myEvent.worldLocation,
									 &co, &tr, &dragPoint, &type))
	{
		if (tr)
		{
			switch (type)
			{
				Trait *ts;
			case controle1:
			case controle2:
				return;
				break;
			case sommet: 
				/* Il faut supprimer : 
					 1) le trait courant
					 2) faire du trait suivant une ligne 
					 */
				ts= NEXT_TRAIT(tr,co);
				if (ts==tr)							/* boucle ? */
				{												/* suppression du contour */
					if (co->previous)
					{
						co->previous->next= 
							co->next;
						if (co->next)
							co->next->previous=
								co->previous;
					}
					else
					{
						Lettre *l;
						int car= SUIT_getInteger(visu, "char");
						l= LaFonte->lettres + car;
						l->contours= co->next;
						if (co->next)
							co->next->previous= NULL;
					}
					free(co);
					free(tr);
					return;
				}
				else
				{
					if (tr-> next && tr->previous)
					{
						tr->previous->next= tr->next;
						tr->next->previous= tr->previous;
					}
					else if (tr->next==NULL)
					{
						co->start= tr->previous->dernier;	/* existe forcment */
						tr->previous->next= NULL;
					}
					else									/* tr->previous== NULL */
					{
						co->debut=tr->next;
						tr->next->previous=NULL;
					}
					ts->type=ligne;
					free(tr);
					fixe_trace(PREVIOUS_POINT(ts,co), ts);
				}
			}
		}
	}
}

void AlignerAPoint(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	Trait *tr;
	enum typepoint type;
	
	if (select_point(visu, myEvent.worldLocation,
									 &co, &tr, &dragPoint, &type))
	{
		if (tr)
		{
			switch (type)
			{
				Trait *ts;
			case controle1:
			case controle2:
				return;
				break;
			case sommet: 
			{
				GP_point v1, v2, vp1;
				double d1, d2, d3;
				ts= NEXT_TRAIT(tr,co);
				if (tr->type != spline ||
						ts->type !=spline)
					return;
				v1.x= ts->controle1.x -tr->dernier.x;
				v1.y= ts->controle1.y -tr->dernier.y;
				v2.x= tr->controle2.x -tr->dernier.x;
				v2.y= tr->controle2.y -tr->dernier.y;
				d1= sqrt(v1.x*v1.x+ v1.y*v1.y);
				d2= sqrt(v2.x*v2.x+ v2.y*v2.y);
				if (d1==0 || d2 == 0)
					return;
				vp1.x= -v2.x/d2 +v1.x/d1;
				vp1.y= -v2.y/d2 +v1.y/d1;
				d3= sqrt(vp1.x*vp1.x+ vp1.y*vp1.y);
				vp1.x= vp1.x/d3;
				vp1.y= vp1.y/d3;
				ts->controle1.x= tr->dernier.x+ 
					vp1.x*d1;
				ts->controle1.y= tr->dernier.y+ 
					vp1.y*d1;
				tr->controle2.x= tr->dernier.x-
					vp1.x*d2;
				tr->controle2.y= tr->dernier.y- 
					vp1.y*d2;
				fixe_trace(tr->dernier,ts);
				if (tr->previous)
					fixe_trace(tr->previous->dernier,tr);
				else
					fixe_trace(co->start,tr);
			}
				break;
			}
		}
	}
}

void ContourToTop(SUIT_object visu, SUIT_event myEvent)
{
	Contour *co;
	Trait *tr;
	enum typepoint type;
	
	if (select_point(visu, myEvent.worldLocation,
									 &co, &tr, &dragPoint, &type))
	{
		Contour *c;
		if (co && co->next)
		{
			if (co->previous== NULL)
			{
				int car;
				Lettre *l;
				car= SUIT_getInteger(visu, "char");
				l= LaFonte->lettres+ car;
				l->contours= co->next;
				co->next->previous= NULL;
			}
			else
			{
				co->previous->next= co->next;
				co->next->previous= co->previous;
			}
			c= co->next;
			while (c->next)
			{
				c= c->next;
			}
			c->next= co;
			co->previous= c;
			co->next= NULL;
		}
	}
}

void FaireZoom(SUIT_object visu, SUIT_event myEvent)
{
	double delta;
	SUIT_window win;
	GP_point lepoint;
	dragPoint= myEvent.worldLocation;
	
	lepoint= GP_unMapPoint(SUIT_drag(DrawZoom));
	delta= fabs(lepoint.x-dragPoint.x);
	win.bottom_left.x= dragPoint.x-delta;
	win.bottom_left.y= dragPoint.y-delta;
	win.top_right.x= dragPoint.x +delta;
	win.top_right.y= dragPoint.y +delta;
	SUIT_setWindow(visu, "window",win);
}

void HitVisu(SUIT_object visu, SUIT_event myEvent)
{
	if (myEvent.type== MOUSE_DOWN)
	{
		if (Zoom)
		{
			FaireZoom(visu, myEvent);
			Zoom= FALSE;
			SUIT_redisplayRequired(visu);
			return;
		}
		switch (MonEtat)
		{
		case movepoint :
			MovePoint(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case addpoint :
			AddPoint(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case spline2line :
			Spline2Line(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case alignerspline :
			AlignerAPoint(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case deletepoint:
			DetruirePoint(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case addcontour:
			AddContour(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case invertcontour:
			InvertContour(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case contour2top:
			ContourToTop(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case bbmove:
			BBMove(visu,myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case setruleorigin:
			SUIT_setDouble(visu, "rule origine x", myEvent.worldLocation.x);
			SUIT_setDouble(visu, "rule origine y", myEvent.worldLocation.y);
			SUIT_redisplayRequired(visu);
			break;
		case setruleticks:
			SUIT_setDouble(visu, "rule tick x", myEvent.worldLocation.x);
			SUIT_setDouble(visu, "rule tick y", myEvent.worldLocation.y);
			SUIT_redisplayRequired(visu);
			break;
		case selectcontour:
			SelectContour(visu,myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case deletecontour:
			DeleteContour(visu,myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case moveselected:
			MoveSelected(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case rotateselected:
			RotateSelected(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		case scaleselected:
			ScaleSelected(visu, myEvent);
			SUIT_redisplayRequired(visu);
			break;
		}
	}
}

void Paintbb(SUIT_object visu, Lettre* l)
{
	if (SUIT_getBoolean(visu, "show bb"))
	{
		SUIT_window win;
		GP_setColor(GP_defColor("blue", TRUE));			
		GP_rectangleCoord(l->llx,l->lly,l->urx,l->ury);
		GP_lineCoord(l->llx,l->lly,l->llx+l->width,l->lly);
		win= OBJECT_WINDOW(visu);
		GP_setColor(GP_defColor("antiquewhite", TRUE));					
		GP_lineCoord(win.bottom_left.x,0.0,win.top_right.x,0.0);
		GP_lineCoord(0.0,win.bottom_left.y,0.0,win.top_right.y);
		
	}
}


void PaintContour(SUIT_object visu, Contour* c)
{
	Trait *t;
	static GP_point *pts= NULL;
	static int available=0;
	const increment=10*(MAXPTS+1);
	int nbr_pts;
	
	if (pts== NULL)
	{
		pts= (GP_point *) malloc(increment*sizeof(GP_point));
		available= increment;
	}
	/* nombre de points necessaires ? */
	nbr_pts= 1;										/* origine */
	if (c->debut!=NULL) 
	{
		for (t=c->debut; t != NULL; t= t->next)
		{
			switch (t->type)
			{
			case ligne:
				nbr_pts++;
				break;
			case spline:
				nbr_pts+= MAXPTS+1;
			}
		}
		/* l'origine apparait 2 fois */
		if (available < nbr_pts)
		{
			pts= (GP_point *)
				realloc(pts,
								(available= 
								 (nbr_pts/increment + 1)* increment)
								* sizeof(GP_point));
			/* en rajoute trop si la division est juste */
		}
		/* maintenant, on cree le contour : */
		nbr_pts= 0;
		pts[nbr_pts++]= c->start;			
		for (t=c->debut; t != NULL; t= t->next)
		{
			switch (t->type)
			{
			case ligne:
				if (SUIT_getBoolean(visu, "Show points"))
				{
					GP_setColor(GP_defColor("green", TRUE));
					GP_marker(pts[nbr_pts-1]);
				}
				pts[nbr_pts++]= t->dernier;
				if (SUIT_getBoolean(visu, "Show points"))
					GP_marker(pts[nbr_pts-1]);
				break;
			case spline:
			{
				int i,num;
				num= nbr_pts-1;
				for (i=0; i< MAXPTS; i++)
					pts[nbr_pts++]= t->trace[i];
				pts[nbr_pts++]= t->dernier;
				if (SUIT_getBoolean(visu, "Show points"))
				{
					GP_setColor(GP_defColor("green", TRUE));
					GP_line(pts[num], t->controle1);
					GP_line(t->controle2, t->dernier);
					GP_marker(pts[num]);
					GP_marker(t->controle1);
					GP_marker(t->controle2);
					GP_marker(t->dernier);
					GP_setColor(GP_defColor("blue", TRUE));
					GP_line(t->controle1, t->controle2);
				}
			}
				break;
			}
		}
	}
	if (SUIT_getBoolean(visu, "preview"))
	{
		if (c->type == positif) 
			GP_setColor(GP_defColor("black", TRUE));
		else
			GP_setColor(SUIT_getColor(visu,BACKGROUND_COLOR));
		nbr_pts--;
		if (nbr_pts > 2)
			GP_fillPolygon(nbr_pts, pts);
	}
	else
	{
		if (c->selected)
			GP_setColor(GP_defColor("yellow", TRUE));
		else
			if (c->type == positif) 
				GP_setColor(GP_defColor("black", TRUE));
			else
				GP_setColor(GP_defColor("red", TRUE));
		nbr_pts--;
		if (nbr_pts > 2)
			GP_polygon(nbr_pts, pts);
		else
			if (nbr_pts > 1)
				GP_polyLine(nbr_pts, pts);
			else
				if (nbr_pts == 1)
					GP_marker(*pts);
	}
}

void PaintSelected(SUIT_object visu, Contour* c)
{
	Trait *t;
	GP_point pts[MAXPTS+2];
	GP_point last;
	
	last= c->start;
	GP_setColor(GP_defColor("yellow", TRUE));
	
	for (t=c->debut; t != NULL; t= t->next)
	{
		if (t->selected)
		{
			int i;
			switch (t->type)
			{
			case ligne:
				GP_line(last,t->dernier);
				break;
			case spline:
				pts[0]= last;
				for (i=1; i<= MAXPTS; i++)
					pts[i]= t->trace[i-1];
				pts[MAXPTS+1]= t->dernier;
				GP_polyLine(MAXPTS+2, pts);
				break;
			}
		}
		last = t->dernier;
	}
}

void PaintLettre(SUIT_object visu, Lettre* l)
{
	Contour *c;
	if (l->used)
	{
		c= l->contours;
		
		while (c != NULL)
		{
			PaintContour(visu,c);
			PaintSelected(visu,c);
			c= c->next;
		}
	}
}

boolean IsVisible(GP_point p, SUIT_window w)
{
	return 
		(p.x >= w.bottom_left.x && 
		 p.y >= w.bottom_left.y && 
		 p.x <= w.top_right.x &&
		 p.y <= w.top_right.y);
}

void PaintRule(SUIT_object visu)
{
	GP_point origine, tickpoint;
	GP_point vect, m;
	SUIT_window win;
	GP_setMarkerStyle(MARKER_X);
	win= OBJECT_WINDOW(visu);
	
	origine.x= SUIT_getDouble(visu, "rule origine x");
	origine.y= SUIT_getDouble(visu, "rule origine y");
	tickpoint.x= SUIT_getDouble(visu, "rule tick x");
	tickpoint.y= SUIT_getDouble(visu, "rule tick y");
	GP_setColor(GP_defColor("azure", TRUE));
	
	GP_line(origine, tickpoint);
	vect.x = tickpoint.x -origine.x;
	vect.y = tickpoint.y -origine.y;
	m= origine;
	while (IsVisible(m, win))
	{
		if (SUIT_getBoolean(visu, "Show ticks"))
			GP_marker(m);
		m.x+= vect.x;
		m.y+= vect.y;
	}
	GP_line(origine, m);
	m= origine;
	while (IsVisible(m, win))
	{
		if (SUIT_getBoolean(visu, "Show ticks"))
			GP_marker(m);
		m.x-= vect.x;
		m.y-= vect.y;
	}
	GP_line(origine, m);
}



void PaintVisu(SUIT_object visu)
{
	int car;
	car= SUIT_getInteger(visu, "char");
	if (LaFonte)
	{
		Paintbb(visu, LaFonte->lettres + car);
		PaintLettre(visu, LaFonte->lettres + car);
		if (SUIT_getBoolean(visu, "Show rule"))
		{
			PaintRule(visu);
		}
	}
}

SUIT_object CreateVisu (char*name)
{
	SUIT_object visu;
	visu= SUIT_createObject(name, "Visu class");
  SUIT_addDisplayToObject (visu, "standard", 
													 HitVisu, PaintVisu);	 /* 4 */
	SUIT_registerClass ("Visu class", CreateVisu, 
											"This class of widgets displays a BZR font"); /* 5 */
	return (visu);					 
}

void LoadFont(SUIT_object bouton)
{
	char * fname, str[MAXPATHLEN];
	
	fname = SUIT_askForFileName((getwd(str)),
															"load","File name:");
	if (fname)
	{
		int i;
		/* destruction de la precedente */
		for (i=0; i< CHRINFONT; i++)
			if (LaFonte->lettres[i].used)
				DetruireContours(LaFonte->lettres[i].contours);
		free(LaFonte);
		
		LaFonte= read_font(fname);
	}
	SUIT_redisplayRequired(LaVisu);
}

void WriteFont(SUIT_object bouton)
{
	char * fname, str[MAXPATHLEN];
	
	fname = SUIT_askForFileName((getwd(str)),
															"save","File name:");
	if (fname)
		write_fonte(LaFonte, fname);
}

void SaveFont(SUIT_object bouton)
{
	write_fonte(LaFonte, LaFonte->filename);
}

void IncreaseNumber(SUIT_object bouton)
{
	int car;
	static char buf[10];
	car= SUIT_getInteger(LaVisu, "char");

	if (car <255)
		car++;
	SUIT_setInteger(LaVisu, "char", car);
	sprintf(buf, "%i", car);
	SUIT_setText(Affiche, CURRENT_VALUE, buf);
	SUIT_redisplayRequired(LaVisu);	
	SUIT_redisplayRequired(Affiche);
}

void DecreaseNumber(SUIT_object bouton)
{
	int car;
	static char buf[10];
	car= SUIT_getInteger(LaVisu, "char");
	if (car>0)
		car--;
	SUIT_setInteger(LaVisu, "char", car);
	sprintf(buf, "%i", car);
	SUIT_setText(Affiche, CURRENT_VALUE, buf);
	SUIT_redisplayRequired(LaVisu);	
	SUIT_redisplayRequired(Affiche);
}



void ChooseNumber(SUIT_object bouton)
{
	int car;
	car= atoi(SUIT_getText(bouton, CURRENT_VALUE));
	SUIT_setInteger(LaVisu, "char", car);
	SUIT_redisplayRequired(LaVisu);
}

void ZoomInButton(SUIT_object bouton)
{
	Zoom=TRUE;
}

void ZoomOutButton(SUIT_object bouton)
{
	SUIT_window win;
	GP_point centre;
	double h,l;
	const double ratio= 2.0;
	
	win= OBJECT_WINDOW(LaVisu);
	h=  win.top_right.y-win.bottom_left.y;
	l=  win.top_right.x-win.bottom_left.x;
	centre.x= (win.top_right.x+win.bottom_left.x)/2.0;
	centre.y= (win.top_right.y+win.bottom_left.y)/2.0;
	win.bottom_left.x= centre.x- l*ratio/2.0;
	win.bottom_left.y= centre.y- h*ratio/2.0;
	win.top_right.x= centre.x+ l*ratio/2.0;
	win.top_right.y= centre.y+ h*ratio/2.0;
	SUIT_setWindow(LaVisu,"window",win);
	SUIT_redisplayRequired(LaVisu);
}

void RuleHorizontal(SUIT_object bouton)
{
	GP_point origine, tickpoint;
	double d;
	origine.x= SUIT_getDouble(LaVisu, "rule origine x");
	origine.y= SUIT_getDouble(LaVisu, "rule origine y");
	tickpoint.x= SUIT_getDouble(LaVisu, "rule tick x");
	tickpoint.y= SUIT_getDouble(LaVisu, "rule tick y");
	d= DIST(origine,tickpoint);
	if (d !=0.0)
	{
		SUIT_setDouble(LaVisu, "rule tick x", origine.x + sqrt(d));
		SUIT_setDouble(LaVisu, "rule tick y", origine.y);
	}
	else
	{
		SUIT_setDouble(LaVisu, "rule tick x", origine.x + 1.0);
		SUIT_setDouble(LaVisu, "rule tick y", origine.y);
	}
	if (SUIT_getBoolean(LaVisu, "Show rule"))
		SUIT_redisplayRequired(LaVisu);
}


void RuleVertical(SUIT_object bouton)
{
	GP_point origine, tickpoint;
	double d;
	origine.x= SUIT_getDouble(LaVisu, "rule origine x");
	origine.y= SUIT_getDouble(LaVisu, "rule origine y");
	tickpoint.x= SUIT_getDouble(LaVisu, "rule tick x");
	tickpoint.y= SUIT_getDouble(LaVisu, "rule tick y");
	d= DIST(origine,tickpoint);
	if (d !=0.0)
	{
		SUIT_setDouble(LaVisu, "rule tick x", origine.x);
		SUIT_setDouble(LaVisu, "rule tick y", origine.y+sqrt(d));
	}
	else
	{
		SUIT_setDouble(LaVisu, "rule tick x", origine.x);
		SUIT_setDouble(LaVisu, "rule tick y", origine.y+1.0);
	}
	if (SUIT_getBoolean(LaVisu, "Show rule"))
		SUIT_redisplayRequired(LaVisu);
}

void HalfRule(SUIT_object bouton)
{
	GP_point origine, tickpoint, middle;
	origine.x= SUIT_getDouble(LaVisu, "rule origine x");
	origine.y= SUIT_getDouble(LaVisu, "rule origine y");
	tickpoint.x= SUIT_getDouble(LaVisu, "rule tick x");
	tickpoint.y= SUIT_getDouble(LaVisu, "rule tick y");
	MOYENNE(middle,origine,tickpoint);
	tickpoint= middle;
	SUIT_setDouble(LaVisu, "rule tick x", tickpoint.x);
	SUIT_setDouble(LaVisu, "rule tick y", tickpoint.y);
	if (SUIT_getBoolean(LaVisu, "Show rule"))
		SUIT_redisplayRequired(LaVisu);
}

void ChooseRadioButton(SUIT_object radio_buttons)
{
	char *curr_choice;
	int i=0;
	curr_choice = SUIT_getEnumString(radio_buttons, CURRENT_VALUE);
	while (strcmp(curr_choice, etats_str[i])!= 0)
		i++;
	MonEtat= i;
	SUIT_setBoolean(LaVisu, "Show points", TRUE);
}

void CopyToClip(SUIT_object bouton)
{
	CopySelected();
}

void GetFromClip(SUIT_object bouton)
{
	pasteClip();
	SUIT_redisplayRequired(LaVisu);
}

void EmptyClip(SUIT_object bouton)
{
	DetruireContours(clip);
	clip= NULL;
}

void NewLetter(SUIT_object bouton)
{
	int car;
	Lettre *l;
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	l->used= TRUE;
	l->contours= NULL;
	l->code= car;
	l->urx=l->ury= l->width= 14.0;
	l->llx= l->lly= 0.0;
	SUIT_redisplayRequired(LaVisu);
}

void DeleteLetter(SUIT_object bouton)
{
	int car;
	Lettre *l;
	
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres + car;
	if (l->used)
	{
		l->used= FALSE;
		DetruireContours(l->contours);
		SUIT_redisplayRequired(LaVisu);
	}
}


void MakeExchange(SUIT_object type)
{
	unsigned int newcar, car;
	Lettre l;
	newcar= atoi(SUIT_getText(type, CURRENT_VALUE));
	car= SUIT_getInteger(LaVisu, "char");
	l= LaFonte->lettres[car];
	LaFonte->lettres[car]= LaFonte->lettres[newcar];
	LaFonte->lettres[newcar]= l;	
	SUIT_redisplayRequired(LaVisu);
}


int main(int argc, char ** argv)
{
	int i;
	SUIT_object buttons;
	
	SUIT_initFromCode(argv[0]);
	/* SUIT_init(argv[0]);*/

	LaFonte= NEW(Font);
	LaFonte->size= 14.0;
	for (i=0; i< 256; i++)
		LaFonte->lettres[i].used= FALSE;
	LaFonte->name= "unknown";

	SUIT_createLabel("@b(Sbzr)");
	SUIT_createLabel("@i(A software for editing BZR files)");
	SUIT_createLabel("@(copyright) 1994  Serge Rosmorduc");
	SUIT_createTextBox("discl", disclaimer);
	SUIT_createLabel("@b(thanks to)  Ronan Corler and Philippe Louet");

	SUIT_createButton("Load Font", LoadFont);	
	SUIT_createButton("Write Font as", WriteFont);	
	SUIT_createButton("Save Font", SaveFont);	
	
	SUIT_createButton("Zoom in", ZoomInButton);	
	SUIT_createButton("Zoom out", ZoomOutButton);	
	
	SUIT_createButton("set rule horizontal", RuleHorizontal);
	SUIT_createButton("set rule vertical", RuleVertical);
	SUIT_createButton("Half Rule Ticksize", HalfRule);
	
	SUIT_createButton("copy to clip", CopyToClip);
	SUIT_createButton("get from clip", GetFromClip );
	SUIT_createButton("empty clip", EmptyClip);
	SUIT_createButton("new letter", NewLetter);
	SUIT_createButton("delete letter", DeleteLetter);
	SUIT_createButton("Move bb to zero", MoveToZero);
	SUIT_createButton("Compute BB", ComputeBB);

	Affiche= SUIT_createTypeInBox("char number", ChooseNumber);
	SUIT_createTypeInBox("Exchange with", MakeExchange);

	SUIT_createArrowButton("char+", IncreaseNumber);
	SUIT_createArrowButton("char-", DecreaseNumber);
	
	buttons = SUIT_createRadioButtons("choix actions", ChooseRadioButton);/* 1 */
	for (i=0; etats_str[i][0] != '0'; i++)
		SUIT_addButtonToRadioButtons (buttons, etats_str[i]);
	LaVisu= CreateVisu("mavisu");
	SUIT_createDoneButton (NULL);
	SUIT_beginStandardApplication();
	exit(0);
}
